-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local hidePosition =
{
  [ -1] = true, -- World/Cosmic
  [ 13] = true, -- Kalimdor
  [ 14] = true, -- Eastern Kingdoms
  [466] = true, -- Outland
  [485] = true, -- Northrend
  [613] = true, -- Vashj'ir
  [689] = true, -- Stranglethorn Vale
  [751] = true, -- Maelstrom
  [862] = true  -- Pandaria
};

local playerZoneId = -1;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to get the ID of the player's current zone
--
function NigglesLocatorOnEvent(self, event, ...)
  -- Process the event
  if (event == "ZONE_CHANGED_NEW_AREA") then
    -- Local Variables
    local currentMapId = GetCurrentMapAreaID();

    -- Check if the map is displayed
    if (WorldMapFrame:IsShown()) then
      SetMapToCurrentZone();
      playerZoneId = GetCurrentMapAreaID();
      SetMapByID(currentMapId);
    else
      playerZoneId = currentMapId;
    end
  end

  return;
end

--
-- Function to initialise the Locator frame when it is loaded
--
function NigglesLocatorOnLoad(self)
  -- Register the frame for drag operations
  self:RegisterForDrag("LeftButton");

  -- Get the player's current zone
  SetMapToCurrentZone();
  NigglesLocatorOnEvent(self, "ZONE_CHANGED_NEW_AREA");

  -- Initialise the last position displayed
  self.lastX = 0;
  self.lastY = 0;

  return;
end

--
-- Function to toggle the displaying of the Locator frames
--
function NigglesLocatorShow(show)
  -- Work out the correct state for the Locator frames
  if (show ~= nil) then
    NigglesDBC.settings.LocatorShown = show;
  else
    NigglesDBC.settings.LocatorShown = not NigglesLocator:IsShown();
  end

  -- Show/Hide the Locator frame
  NigglesLocator:SetShown(NigglesDBC.settings.LocatorShown);

  -- Create or destroy a timer to update the Locator frame
  if (NigglesLocator:IsShown()) then
    NigglesTimerCreate("Locator", NigglesLocatorUpdate, nil, nil, 0.1, true);
  else
    NigglesTimerDestroy("Locator");
  end

  return;
end

--
-- Function to update the coordinates displayed in the Locator frame
--
function NigglesLocatorUpdate()
  -- Local Variables
  local posX, posY = GetPlayerMapPosition("player");
  local currentMapId = GetCurrentMapAreaID();

  -- Check if the player's position can be displayed, based on current map
  if ((currentMapId == playerZoneId) and
      ((posX ~= 0) and (posY ~= 0))) then
    -- Adjust the player's current position
    posX = math.floor(posX*1000)/10;
    posY = math.floor(posY*1000)/10;

    -- Update the Locator's X position, if required
    if (posX ~= NigglesLocator.lastX) then
      NigglesLocator.posX:SetFormattedText("%2.1f,", posX);
      NigglesLocator.lastX = posX;
    end

    -- Update the Locator's Y position, if required
    if (posY ~= NigglesLocator.lastY) then
      NigglesLocator.posY:SetFormattedText("%2.1f", posY);
      NigglesLocator.lastY = posY;
    end
  elseif ((currentMapId == playerZoneId) and
          ((posX == 0) and (posY == 0)) and
          (IsInInstance()) and
          (WorldMapFrame:IsShown() == nil)) then
    -- Set the map to the current level of the instance's map
    SetMapToCurrentZone();
  elseif ((NigglesLocator.lastX ~= -1) or (NigglesLocator.lastY ~= -1)) then
    -- Display no position because the position is inaccurate
    NigglesLocator.posX:SetText("");
    NigglesLocator.posY:SetText("");
    NigglesLocator.lastX = -1;
    NigglesLocator.lastY = -1;
  end

  return;
end

--
-- Function to enable/disable the Map Location frame
--
function NigglesMapLocationEnable()
  -- Create the Map Location frame, if required
  if ((NigglesMapLocation == nil) and
      (NigglesDBC.settings.UiMapLocation == true)) then
     CreateFrame("FRAME", "NigglesMapLocation", WorldMapFrame,
       "NigglesMapLocationTemplate");
    NigglesMapLocation.lastX = 0;
    NigglesMapLocation.lastY = 0;
  end

  -- Show/Hide the Map Location frame, if required
  if (NigglesMapLocation ~= nil) then
    NigglesMapLocation:SetShown(NigglesDBC.settings.UiMapLocation);
  end

  return;
end

--
-- Function to update the coordinates displayed in the Map Locator frame
--
function NigglesMapLocationOnUpdate(self, elapsed)
  -- Local Variables
   local cursorX = -1;
   local cursorY = -1;

  -- Check the locator is shown
  if (NigglesDBC.settings.UiMapLocation == true) then
    -- Check the mouse is over the map
    if ((WorldMapButton:IsMouseOver()) and
        (hidePosition[GetCurrentMapAreaID()] ~= true)) then
      -- Get the position of the mouse
      local areaName;
      local areaMinLevel;
      local areaMaxLevel;
      local mapCentreX, mapCentreY = WorldMapButton:GetCenter();
      local mapWidth = WorldMapButton:GetWidth();
      local mapHeight = WorldMapButton:GetHeight();
      local mapScale = WorldMapButton:GetEffectiveScale();

      -- Get and convert the cursor's position into a map position
      cursorX, cursorY = GetCursorPosition();
      cursorX = ((cursorX/mapScale)-(mapCentreX-(mapWidth/2)))/mapWidth;
      cursorY = (mapCentreY+(mapHeight/2)-(cursorY/mapScale))/mapHeight;

      -- Get the name and level range of the current area, if any
      areaName, _, _, _, _, _, _, _, areaMinLevel, areaMaxLevel =
        UpdateMapHighlight(cursorX, cursorY);

      -- Check the cursor isn't over another zone
      if ((areaName ~= nil) and
          (((areaMinLevel ~= nil) and (areaMinLevel > 0) and
            (areaMaxLevel ~= nil) and (areaMaxLevel > 0)) or
           (NigglesIsCity(areaName)))) then
        cursorX = -1;
        cursorY = -1;
      end
    end

    -- Check if the cursor position should be displayed
    if ((cursorX >= 0) and (cursorY >= 0)) then
      -- Update the Map Locator's X position, if required
      if (cursorX ~= lastX) then
        lastX = cursorX;
        self.posX:SetFormattedText("%2.01f,", cursorX*100);
      end

      -- Update the Map Locator's Y position, if required
      if (cursorY ~= lastY) then
        lastY = cursorY;
        self.posY:SetFormattedText("%2.01f", cursorY*100);
      end
    elseif ((lastX ~= -1) and (lastY ~= -1)) then
      -- Clear the Map Locator's position
      self.posX:SetText("");
      self.posY:SetText("");
      self.lastX = -1;
      self.lastY = -1;
    end
  end

  return;
end
